.marks {
	border-color: black;
	border-style: solid;
	border-width: 10px;
	border-radius: 0px;
}

modelbutton check
{
	margin-right: 6px;
}

.marks-filter-button
{
	min-width: 12px;
	min-height: 12px;
	padding-left: 0;
	padding-right: 0;
	padding-top: 1px;
	padding-bottom: 1px;
}

.hidden-license {
	opacity: 0;
	min-width: 0;
	min-height: 0;
	margin: 0;
	padding: 0;
}
