    <child>
      <object class="GtkShortcutsSection">
        <property name="visible">1</property>
        <property name="section-name">mouse</property>
        <property name="title" translatable="yes">Mouse</property>
        <property name="max-height">12</property>
        <child>
          <object class="GtkShortcutsGroup">
            <property name="visible">1</property>
            <property name="title" translatable="yes">Mouse</property>
            <child>
              <object class="GtkShortcutsShortcut">
                <property name="visible">1</property>
                <property name="title" translatable="yes">🖱 Left Click						Next image</property>
              </object>
            </child>
            <child>
              <object class="GtkShortcutsShortcut">
                <property name="visible">1</property>
                <property name="title" translatable="yes">🖱  Mouse Wheel down			Next image ₁</property>
              </object>
            </child>
            <child>
              <object class="GtkShortcutsShortcut">
                <property name="visible">1</property>
                <property name="title" translatable="yes">🖱 Middle Click					Previous image</property>
              </object>
            </child>
            <child>
              <object class="GtkShortcutsShortcut">
                <property name="visible">1</property>
                <property name="title" translatable="yes">🖱  Mouse Wheel up				Previous image ₁</property>
              </object>
            </child>
            <child>
              <object class="GtkShortcutsShortcut">
                <property name="visible">1</property>
                <property name="title" translatable="yes">🖱  Left click and drag				Pan image</property>
              </object>
            </child>
            <child>
              <object class="GtkShortcutsShortcut">
                <property name="visible">1</property>
                <property name="title" translatable="yes">🖱  Control + Mouse wheel up		Zoom in</property>
              </object>
            </child>
            <child>
              <object class="GtkShortcutsShortcut">
                <property name="visible">1</property>
                <property name="title" translatable="yes">🖱  Control + Mouse wheel down	Zoom out</property>
              </object>
            </child>
            <child>
              <object class="GtkShortcutsShortcut">
                <property name="visible">1</property>
                <property name="title" translatable="yes">🖱  Mouse button 8				Forward ₂</property>
              </object>
            </child>
            <child>
              <object class="GtkShortcutsShortcut">
                <property name="visible">1</property>
                <property name="title" translatable="yes">🖱  Mouse button 9				Back ₂</property>
              </object>
            </child>
            <child>
              <object class="GtkShortcutsShortcut">
                <property name="visible">1</property>
                <property name="title" translatable="yes">🖱  Control + left click				Smooth scroll</property>
              </object>
            </child>
            <child>
              <object class="GtkShortcutsShortcut">
                <property name="visible">1</property>
                <property name="title" translatable="yes">🖱  Control + Shift + right click				Display additional information</property>
              </object>
            </child>
            <child>
              <object class="GtkShortcutsShortcut">
                <property name="visible">1</property>
                <property name="title" translatable="yes" />
                <property name="subtitle" translatable="yes">1 Default Mouse wheel functions can be altered by changing the Mouse wheel scrolls image option</property>
              </object>
            </child>
            <child>
              <object class="GtkShortcutsShortcut">
                <property name="visible">1</property>
                <property name="title" translatable="yes" />
                <property name="subtitle" translatable="yes">2 Requires a multi-button mouse.</property>
              </object>
            </child>
          </object>
        </child>
      </object>
    </child>
    <child>
      <object class="GtkShortcutsSection">
        <property name="visible">1</property>
        <property name="section-name">arrow-keys</property>
        <property name="title" translatable="yes">Arrow Keys</property>
        <property name="max-height">12</property>
        <child>
          <object class="GtkShortcutsGroup">
            <property name="visible">1</property>
            <property name="title" translatable="yes">Arrows</property>
            <child>
              <object class="GtkShortcutsShortcut">
                <property name="visible">1</property>
                <property name="title" translatable="yes">Pan image</property>
                <property name="accelerator">Up Down Left Right</property>
              </object>
            </child>
            <child>
              <object class="GtkShortcutsShortcut">
                <property name="visible">1</property>
                <property name="title" translatable="yes">Pan image faster</property>
                <property name="accelerator">&lt;shift&gt;Up Down Left Right</property>
              </object>
            </child>
            <child>
              <object class="GtkShortcutsShortcut">
                <property name="visible">1</property>
                <property name="title" translatable="yes">Pan to respective edge of image</property>
                <property name="accelerator">&lt;ctrl&gt;Up Down Left Right</property>
              </object>
            </child>
            <child>
              <object class="GtkShortcutsShortcut">
                <property name="visible">1</property>
                <property name="title" translatable="yes">Move to respective split pane</property>
                <property name="accelerator">&lt;alt&gt;Up Down Left Right</property>
              </object>
            </child>
          </object>
        </child>
      </object>
    </child>
